#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import warnings

from support import InstanceTestCase

from minizinc import Method, Solver, Status, default_driver
from minizinc.error import MiniZincWarning
from minizinc.helpers import check_result, check_solution


class CheckResults(InstanceTestCase):
    code = """
        array[1..2] of var 1..10: x;
        constraint x[1] + 1 = x[2];
    """
    other_solver = Solver.lookup("chuffed")

    def test_correct(self):
        assert self.instance.method == Method.SATISFY
        result = self.instance.solve()
        assert check_result(self.instance, result, self.other_solver)

    def test_incorrect(self):
        assert self.instance.method == Method.SATISFY
        with warnings.catch_warnings(record=True) as w:
            result = self.instance.solve()
            result.solution = self.instance.output_type(x=[2, 1])
            assert not check_result(self.instance, result, self.other_solver)
            if default_driver.parsed_version >= (2, 6, 0):
                assert len(w) == 1
                assert issubclass(w[-1].category, MiniZincWarning)
                assert "model inconsistency" in str(w[-1].message)

    def test_check_all(self):
        assert self.instance.method == Method.SATISFY
        result = self.instance.solve(all_solutions=True)
        assert check_result(
            self.instance, result, self.other_solver, range(len(result.solution))
        )

    def test_check_specific(self):
        assert self.instance.method == Method.SATISFY
        result = self.instance.solve(nr_solutions=5)
        assert check_result(self.instance, result, self.other_solver, [1, 2])

    def test_dict(self):
        assert check_solution(
            self.instance, {"x": [5, 6]}, Status.SATISFIED, self.other_solver
        )

    def test_enum(self):
        self.instance.add_string("""enum Foo = {A, B};var Foo: f;""")
        result = self.instance.solve()
        assert check_result(self.instance, result, self.other_solver)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
import minizinc


def test_default_driver():
    assert minizinc.default_driver is not None


def test_version():
    # Test normal behaviour
    assert "MiniZinc" in minizinc.default_driver.minizinc_version



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
import warnings
from typing import Optional, Type

from ._version import version as __version__
from .driver import Driver, find_driver
from .error import ConfigurationError, MiniZincError
from .instance import Instance as GenInstance
from .model import Method, Model
from .result import Result, Status
from .solver import Solver

Instance: Type[GenInstance]

logger = logging.getLogger("minizinc")

#: Default MiniZinc driver used by the python package
try:
    default_driver: Optional[Driver] = find_driver()
    if default_driver is not None:
        default_driver.make_default()
    else:
        warnings.warn(
            "MiniZinc was not found on the system. No default driver could be "
            "initialised.",
            RuntimeWarning,
        )
except ConfigurationError as err:
    warnings.warn(
        f"The MiniZinc version found on the system is incompatible with "
        f"MiniZinc Python:\n{err}\n No default driver could be initialised.",
        RuntimeWarning,
    )

__all__ = [
    "__version__",
    "default_driver",
    "find_driver",
    "Driver",
    "Instance",
    "Method",
    "MiniZincError",
    "Model",
    "Result",
    "Solver",
    "Status",
]



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest
from support import InstanceTestCase

import minizinc
from minizinc.error import (
    AssertionError,
    EvaluationError,
    MiniZincError,
    SyntaxError,
    TypeError,
)


class AssertionTest(InstanceTestCase):
    code = """
        array [1..10] of int: a = [i | i in 1..10];
        constraint assert(forall (i in 1..9) (a[i] > a[i + 1]), "a not decreasing");
        var 1..10: x;
        constraint a[x] = max(a);
        solve satisfy;
    """

    def test_assertion_error(self):
        with pytest.raises(AssertionError, match="a not decreasing") as error:
            self.instance.solve()
        loc = error.value.location
        assert str(loc.file).endswith(".mzn")
        assert loc.lines == (3, 3)
        if minizinc.default_driver.parsed_version >= (2, 6, 0):
            assert loc.columns == (27, 62)


class TypeErrorTest(InstanceTestCase):
    code = """
        array[1..2] of var int: i;
        constraint i = 1.5;
    """

    def test_type_error(self):
        with pytest.raises(TypeError, match="No matching operator found") as error:
            self.instance.solve()
        loc = error.value.location
        assert str(loc.file).endswith(".mzn")
        assert loc.lines == (3, 3)
        assert loc.columns == (20, 26)


class SyntaxErrorTest(InstanceTestCase):
    code = "constrain true;"

    def test_syntax_error(self):
        with pytest.raises(SyntaxError, match="unexpected bool literal") as error:
            self.instance.solve()
        loc = error.value.location
        assert str(loc.file).endswith(".mzn")
        assert loc.lines == (1, 1)
        assert loc.columns == (11, 14)


class EvaluationErrorTest(InstanceTestCase):
    def test_infinite_recursion(self):
        self.instance.add_string(
            """
test overflow(int: x) = overflow(x + 1);
int: cause_overflow = overflow(1);
"""
        )

        with pytest.raises(
            MiniZincError,
            match="stack overflow"
            if minizinc.default_driver.parsed_version >= (2, 6, 0)
            else "non-zero exit code",
        ):
            self.instance.solve()

    def test_evaluation_error(self):
        self.instance.add_string(
            """
array [1..3] of int: a = [1, 2, 3, 4];

solve satisfy;
"""
        )
        with pytest.raises(EvaluationError, match="index set") as error:
            self.instance.solve()
        loc = error.value.location
        print(loc)
        assert str(loc.file).endswith(".mzn")
        assert loc.lines == (2, 2)
        if minizinc.default_driver.parsed_version >= (2, 6, 0):
            assert loc.columns == (1, 22)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from minizinc.instance import Instance


class APIInstance(Instance):
    pass
    # _ptr: c_void_p

    # def __init__(self, files: Optional[List[Union[Path, str]]] = None):
    #     self._ptr = self.driver._minizinc_instance_init()
    #     if self._ptr is None:
    #         msg = self.driver._minizinc_error()
    #         raise SystemError(msg.decode())
    #     super().__init__(files)

    # def solve(self, solver, *args, **kwargs):
    #     pass

    # def __del__(self):
    #     if self._ptr is not None:
    #         succes = self.driver._minizinc_instance_destroy(self._ptr)
    #         if not succes:
    #             msg = self.driver._minizinc_error()
    #             raise SystemError(msg.decode())

    # def __setitem__(self, key: str, value: Any):
    #     pass

    # def __getitem__(self, key: str) -> Any:
    #     pass

    # def add_to_model(self, code: str) -> None:
    #     pass

    # def branch(self):
    #     pass

    # @property
    # def method(self) -> Method:
    #     pass



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import warnings

import pytest
from support import InstanceTestCase

import minizinc
from minizinc.error import MiniZincWarning
from minizinc.result import Status


class TestAssign(InstanceTestCase):
    code = """
        include "globals.mzn";
        int: n;
        array[1..4] of var 1..5: x;
        constraint increasing(x);
        constraint alldifferent(x);
        constraint sum(x) = n;
    """

    def test_assign(self):
        self.instance["n"] = 14
        result = self.instance.solve(all_solutions=True)
        assert result.status == Status.ALL_SOLUTIONS
        assert len(result.solution) == 1
        assert result[0, "x"] == [i for i in range(2, 5 + 1)]

    def test_reassign(self):
        self.instance["n"] = 14
        with pytest.raises(AssertionError, match="cannot be assigned multiple values."):
            self.instance["n"] = 15


class TestPythonConflict(InstanceTestCase):
    code = """
        include "globals.mzn";
        var 1..2: return;
        constraint return > 1;
    """

    def test_rename(self):
        with pytest.warns(SyntaxWarning):
            result = self.instance.solve()
            assert result.solution.mzn_return == 2


class TestBranch(InstanceTestCase):
    code = """
        include "globals.mzn";
        var 14..15: n;
        array[1..4] of var 1..5: x;
        constraint increasing(x);
        constraint sum(x) = n;
    """

    def test_add_data(self):
        result = self.instance.solve(all_solutions=True)
        assert result.status == Status.ALL_SOLUTIONS
        assert len(result.solution) == 12
        with self.instance.branch() as child:
            child["n"] = 15
            result = child.solve(all_solutions=True)
            assert result.status == Status.ALL_SOLUTIONS
            assert len(result.solution) == 5
        with self.instance.branch() as child:
            child["n"] = 14
            result = child.solve(all_solutions=True)
            assert result.status == Status.ALL_SOLUTIONS
            assert len(result.solution) == 7

    def test_extra_constraint(self):
        self.instance["n"] = 14
        result = self.instance.solve(all_solutions=True)
        assert result.status == Status.ALL_SOLUTIONS
        assert len(result.solution) == 7
        with self.instance.branch() as child:
            child.add_string("constraint all_different(x);")
            result = child.solve(all_solutions=True)
            assert result.status == Status.ALL_SOLUTIONS
            assert len(result.solution) == 1

    def test_replace_data(self):
        self.instance["n"] = 14
        result = self.instance.solve(all_solutions=True)
        assert result.status == Status.ALL_SOLUTIONS
        assert len(result.solution) == 7
        with self.instance.branch() as child:
            child["n"] = 15
            with warnings.catch_warnings(record=True) as w:
                result = child.solve(all_solutions=True)
                assert result.status == Status.UNSATISFIABLE
                assert len(result.solution) == 0
                if minizinc.default_driver.parsed_version >= (2, 6, 0):
                    assert len(w) == 1
                    assert issubclass(w[-1].category, MiniZincWarning)
                    assert "model inconsistency" in str(w[-1].message)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest

from minizinc import Instance, Solver
from minizinc.instance import Instance as GenInstance


class InstanceTestCase(unittest.TestCase):
    code = ""
    instance: GenInstance
    solver: Solver

    def setUp(self):
        self.solver = Solver.lookup("gecode")
        self.instance = Instance(self.solver)
        self.instance.add_string(self.code)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
import enum

import pytest
from support import InstanceTestCase

from minizinc import Instance
from minizinc.result import Status


class TestEnum(InstanceTestCase):
    code = """
    enum DAY = {Mo, Tu, We, Th, Fr, Sa, Su};
    var DAY: d;
    """

    def test_value(self):
        self.instance.add_string("constraint d == Mo;")
        result = self.instance.solve()
        assert isinstance(result["d"], str)
        assert result["d"] == "Mo"

    def test_cmp_in_instance(self):
        self.instance.add_string("var DAY: d2;")
        self.instance.add_string("constraint d < d2;")
        result = self.instance.solve()
        assert isinstance(result["d"], str)
        assert isinstance(result["d2"], str)
        # TODO: assert result["d"] < result["d2"]

    def test_cmp_between_instances(self):
        append = "constraint d == Mo;"
        self.instance.add_string(append)
        result = self.instance.solve()

        inst = Instance(self.solver)
        inst.add_string(self.code + append)
        result2 = inst.solve()
        assert isinstance(result["d"], str)
        assert isinstance(result2["d"], str)
        assert result["d"] == result2["d"]

        inst = Instance(self.solver)
        inst.add_string(
            """
            enum DAY = {Mo, Tu, We, Th, Fr};
            var DAY: d;
            """
            + append
        )
        result2 = inst.solve()
        assert result["d"] == result2["d"]

    def test_assign(self):
        self.instance = Instance(self.solver)
        self.instance.add_string(
            """
            enum TT;
            var TT: t1;
            """
        )
        TT = enum.Enum("TT", ["one"])
        self.instance["TT"] = TT
        result = self.instance.solve()

        assert isinstance(result["t1"], TT)
        assert result["t1"] is TT.one

    def test_collections(self):
        self.instance = Instance(self.solver)
        self.instance.add_string(
            """
            enum TT;
            array[int] of var TT: arr_t;
            var set of TT: set_t;
            """
        )
        TT = enum.Enum("TT", ["one", "two", "three"])
        self.instance["TT"] = TT
        self.instance["arr_t"] = [TT(3), TT(2), TT(1)]
        self.instance["set_t"] = {TT(2), TT(1)}
        result = self.instance.solve()

        assert result["arr_t"] == [TT(3), TT(2), TT(1)]
        assert result["set_t"] == {TT(1), TT(2)}

    def test_intenum_collections(self):
        self.instance = Instance(self.solver)
        self.instance.add_string(
            """
            enum TT;
            % array[int] of var TT: arr_t;
            var set of TT: set_t;
            """
        )
        TT = enum.IntEnum("TT", ["one", "two", "three"])
        self.instance["TT"] = TT
        # TODO:  self.instance["arr_t"] = [TT(3), TT(2), TT(1)]
        self.instance["set_t"] = {TT(2), TT(1)}
        result = self.instance.solve()

        # TODO: assert result["arr_t"] == [TT(3), TT(2), TT(1)]
        assert result["set_t"] == {TT(1), TT(2)}


class TestSets(InstanceTestCase):
    def test_ranges(self):
        self.instance.add_string(
            """
            var set of 0..10: s;
            set of int: s1;
            constraint s1 = s;
            """
        )

        self.instance["s1"] = range(1, 4)
        result = self.instance.solve()
        assert isinstance(result["s"], range)
        assert result["s"] == range(1, 4)


class TestString(InstanceTestCase):
    code = """
    array[int] of string: names;
    var index_set(names): x;
    string: name ::output_only ::add_to_output = names[fix(x)];
    """

    def test_string(self):
        names = ["Guido", "Peter"]
        self.instance["names"] = names

        result = self.instance.solve()
        assert result.solution.name in names


class TestNumPy(InstanceTestCase):
    def test_nparray_bool(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("array[int] of bool: x;")
        self.instance["x"] = numpy.array([True, False], dtype=numpy.bool_)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_nparray_f32(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("array[int] of float: x;")
        self.instance["x"] = numpy.array([1, 2, 3], dtype=numpy.float32)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_nparray_f64(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("array[int] of float: x;")
        self.instance["x"] = numpy.array([1, 2, 3], dtype=numpy.float64)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_nparray_int32(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("array[int] of int: x;")
        self.instance["x"] = numpy.array([1, 2, 3], dtype=numpy.int32)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_nparray_int64(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("array[int] of int: x;")
        self.instance["x"] = numpy.array([1, 2, 3], dtype=numpy.int64)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_nparray_uint32(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("array[int] of int: x;")
        self.instance["x"] = numpy.array([1, 2, 3], dtype=numpy.uint32)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_nparray_uint64(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("array[int] of int: x;")
        self.instance["x"] = numpy.array([1, 2, 3], dtype=numpy.uint64)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_npbool(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("bool: x;")
        self.instance["x"] = numpy.bool_(0)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_npf32(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("float: x;")
        self.instance["x"] = numpy.float32(0)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_npf64(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("float: x;")
        self.instance["x"] = numpy.float64(0)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_npint32(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("int: x;")
        self.instance["x"] = numpy.int32(0)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED

    def test_npint64(self):
        numpy = pytest.importorskip("numpy")
        self.instance.add_string("int: x;")
        self.instance["x"] = numpy.int64(0)
        result = self.instance.solve()
        assert result.status is Status.SATISFIED


# Enable on MiniZinc 2.6+ (JSON output of annotations is broken before then)
# class TestAnn(InstanceTestCase):
#     def test_ann_atom(self):
#         self.instance.add_string("ann: x :: add_to_output = promise_total;")
#         result = self.instance.solve()
#         assert result.status is Status.SATISFIED
#         assert result["x"] == "promise_total"

#     def test_ann_call(self):
#         self.instance.add_string(
#         'ann: x :: add_to_output = expression_name("test");')
#         result = self.instance.solve()
#         assert result.status is Status.SATISFIED
#         assert result["x"] == 'expression_name("test")'



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import platform
import shutil
from abc import ABC, abstractmethod
from ctypes.util import find_library
from pathlib import Path
from typing import List, Optional, Type

from minizinc.instance import Instance


class Driver(ABC):
    """The abstract representation of a MiniZinc driver within MiniZinc Python."""

    Solver: Type
    Instance: Instance

    @abstractmethod
    def make_default(self) -> None:
        """Method to override the current default MiniZinc Python driver with the
        current driver.
        """
        pass

    @abstractmethod
    def __init__(self):
        """Creates a new MiniZinc driver

        Raises:
            ConfigurationError: If an the driver version is found to be
                incompatible with MiniZinc Python
        """
        pass

    @property
    @abstractmethod
    def minizinc_version(self) -> str:
        """Reports the version of the MiniZinc Driver

        Report the full version of MiniZinc as reported by the driver,
        including the driver name, the semantic version, the build reference,
        and its authors.

        Returns:
            str: the version of as reported by the MiniZinc driver

        """
        pass

    @abstractmethod
    def available_solvers(self, refresh=False):
        """Returns a list of available solvers

        This method returns the list of solvers available to the Driver object
        according to the current environment. Note that the list of solvers might
        be cached for future usage. The refresh argument can be used to ignore
        the current cache.

        Args:
            refresh (bool): When set to true, the Driver will rediscover the
                available solvers from the current environment.

        Returns:
            Dict[str, List[Solver]]: A dictionary that maps solver tags to MiniZinc
                solver configurations that can be used with the Driver object.
        """
        pass


def find_driver(
    path: Optional[List[str]] = None, name: str = "minizinc"
) -> Optional[Driver]:
    """Finds MiniZinc Driver on default or specified path.

    Find driver will look for the the MiniZinc API or the MiniZinc executable
    to create a Driver for MiniZinc Python. If no path is specified, then the
    paths given by the environment variables appended by MiniZinc's default
    locations will be tried.

    Args:
        path: List of locations to search.
        name: Name of the API or executable.

    Returns:
        Optional[Driver]: Returns a Driver object when found or None.

    """
    driver: Optional[Driver] = None
    if path is None:
        path_bin = os.environ.get("PATH", "").split(os.pathsep)
        path_lib = os.environ.get("LD_LIBRARY_PATH", "").split(os.pathsep)
        path_lib.extend(os.environ.get("DYLD_LIBRARY_PATH", "").split(os.pathsep))
        # Add default MiniZinc locations to the path
        if platform.system() == "Darwin":
            MAC_LOCATIONS = [
                str(Path("/Applications/MiniZincIDE.app/Contents/Resources")),
                str(
                    Path(
                        "~/Applications/MiniZincIDE.app/Contents/Resources"
                    ).expanduser()
                ),
            ]
            path_bin.extend(MAC_LOCATIONS)
            path_lib.extend(MAC_LOCATIONS)
        elif platform.system() == "Windows":
            WIN_LOCATIONS = [
                str(Path("c:/Program Files/MiniZinc")),
                str(Path("c:/Program Files/MiniZinc IDE (bundled)")),
                str(Path("c:/Program Files (x86)/MiniZinc")),
                str(Path("c:/Program Files (x86)/MiniZinc IDE (bundled)")),
            ]
            path_bin.extend(WIN_LOCATIONS)
            path_lib.extend(WIN_LOCATIONS)
    else:
        path_bin = path
        path_lib = path

    path_bin_list = os.pathsep.join(path_bin)
    path_lib_list = os.pathsep.join(path_lib)

    # Try to load the MiniZinc C API
    env_backup = os.environ.copy()
    os.environ["LD_LIBRARY_PATH"] = path_lib_list
    os.environ["DYLD_LIBRARY_PATH"] = path_lib_list
    lib = find_library(name)
    os.environ.clear()
    os.environ.update(env_backup)
    if lib and Path(lib).suffix in [".dll", ".dylib", ".so"]:
        pass
        # TODO:
        # from minizinc.API import APIDriver

        # library = cdll.LoadLibrary(lib)
        # driver = APIDriver(library)
    else:
        # Try to locate the MiniZinc executable
        executable = shutil.which(name, path=path_bin_list)
        if executable is not None:
            from minizinc.CLI import CLIDriver

            driver = CLIDriver(Path(executable))

    return driver



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import asyncio
import contextlib
import json
import os
import re
import sys
import tempfile
import warnings
from dataclasses import field, make_dataclass
from datetime import timedelta
from enum import EnumMeta
from keyword import iskeyword
from numbers import Number
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, cast

import minizinc
from minizinc.error import MiniZincError, parse_error
from minizinc.instance import Instance
from minizinc.json import (
    MZNJSONDecoder,
    MZNJSONEncoder,
    decode_async_json_stream,
    decode_json_stream,
)
from minizinc.model import Method, Model, ParPath, UnknownExpression
from minizinc.result import Result, Status, parse_solution, set_stat
from minizinc.solver import Solver

from .driver import CLIDriver, to_python_type

if sys.version_info >= (3, 8):
    from typing import Final

    SEPARATOR: Final[bytes] = str.encode("----------" + os.linesep)
else:
    SEPARATOR: bytes = str.encode("----------" + os.linesep)


class _GeneratedSolution:
    pass


class CLIInstance(Instance):
    _driver: CLIDriver
    _solver: Solver
    _input: Optional[Dict[str, Type]] = None
    _output: Optional[Dict[str, Type]] = None
    _method: Optional[Method] = None
    _parent: Optional["CLIInstance"] = None
    _field_renames: List[Tuple[str, str]]

    def __init__(
        self,
        solver: Solver,
        model: Optional[Model] = None,
        driver: Optional[CLIDriver] = None,
    ):
        super().__init__(solver, model)
        self._solver = solver
        self._field_renames = []
        if driver is not None:
            self._driver = driver
        elif minizinc.default_driver is not None and isinstance(
            minizinc.default_driver, CLIDriver
        ):
            self._driver = cast(CLIDriver, minizinc.default_driver)
        else:
            raise Exception("No compatible driver provided")

        if model is not None:
            self.output_type = model.output_type
            self._includes = model._includes.copy()
            self._code_fragments = model._code_fragments.copy()
            self._data = dict.copy(model._data)
            self._enum_map = dict.copy(model._enum_map)
            self._checker = True

            # Generate output_type to ensure the same type between different
            # instances of the same model
            if self.output_type is None:
                self.analyse()
                model.output_type = self.output_type

    @contextlib.contextmanager
    def branch(self) -> Iterator[Instance]:  # TODO: Self reference
        child = self.__class__(self._solver)
        child._parent = self

        # Copy current information from analysis
        child._method = self.method
        child.output_type = self.output_type
        child._output = self._output
        child._input = self.input

        with self._lock:
            yield child

    @property
    def method(self) -> Method:
        if self._method is None:
            self.analyse()
        assert self._method is not None
        return self._method

    @contextlib.contextmanager
    def files(self) -> Iterator[List[Path]]:
        """Gets list of files of the Instance

        Files will create a list of paths to the files that together form the
        Instance. Parts of the Instance might be saved to files and are only
        guaranteed to exist while within the created context.

        Yields:
            List of Path objects to existing and created files

        """

        files: List[Path] = []
        fragments: List[str] = []
        data: Dict[str, Any] = {}

        inst: Optional["CLIInstance"] = self
        while inst is not None:
            for k, v in inst._data.items():
                if isinstance(v, UnknownExpression) or k in data:
                    fragments.append(f"{k} = {v};\n")
                elif isinstance(v, EnumMeta):
                    fragments.append(
                        f"{k} = {{{', '.join([i for i in v.__members__])}}};\n"
                    )
                else:
                    data[k] = v
            fragments.extend(inst._code_fragments)
            files.extend(inst._includes)

            inst = inst._parent

        gen_files = []
        try:
            if len(data) > 0:
                file = tempfile.NamedTemporaryFile(
                    prefix="mzn_data", suffix=".json", delete=False
                )
                gen_files.append(file)
                file.write(json.dumps(data, cls=MZNJSONEncoder).encode())
                file.close()
                files.append(Path(file.name))
            if len(fragments) > 0 or len(files) == 0:
                file = tempfile.NamedTemporaryFile(
                    prefix="mzn_fragment", suffix=".mzn", delete=False
                )
                gen_files.append(file)
                for code in fragments:
                    file.write(code.encode())
                file.close()
                files.append(Path(file.name))
            yield files
        finally:
            for file in gen_files:
                file.close()
                os.remove(file.name)

    @property
    def input(self):
        if self._input is None or self._method is None:
            self.analyse()
        return self._input

    @property
    def output(self):
        if self._output is None or self._method is None:
            self.analyse()
        return self._output

    def analyse(self):
        """Discovers basic information about a CLIInstance

        Analyses a given instance and discovers basic information about set
        model such as the solving method, the input parameters, and the output
        parameters. The information found will be stored among the attributes
        of the instance.
        """
        with self.files() as files:
            assert len(files) > 0
            output = self._driver.run(["--model-interface-only"] + files, self._solver)
        if self._driver.parsed_version >= (2, 6, 0):
            interface = None
            for obj in decode_json_stream(output.stdout):
                if obj["type"] == "interface":
                    interface = obj
                    break
        else:
            interface = json.loads(output.stdout)
        old_method = self._method
        self._method = Method.from_string(interface["method"])
        self._input = {}
        for key, value in interface["input"].items():
            self._input[key] = to_python_type(value)
        old_output = self._output
        self._output = {}
        for (key, value) in interface["output"].items():
            self._output[key] = to_python_type(value)
        if interface.get("has_output_item", True):
            self._output["_output_item"] = str
        if self._checker:
            self._output["_checker"] = str

        if self.output_type is None or (
            issubclass(self.output_type, _GeneratedSolution)
            and (self._output != old_output or self._method != old_method)
        ):
            fields = []
            self._field_renames = []
            if self._method is not Method.SATISFY and "objective" not in self._output:
                fields.append(("objective", Number))
            for k, v in self._output.items():
                if k in ["_output_item", "_checker"]:
                    fields.append((k, str, field(default="")))
                elif iskeyword(k):
                    warnings.warn(
                        f"MiniZinc field '{k}' is a Python keyword. It has been "
                        f"renamed to 'mzn_{k}'",
                        SyntaxWarning,
                    )
                    self._field_renames.append((k, "mzn_" + k))
                    fields.append(("mzn_" + k, v))
                else:
                    fields.append((k, v))

            minizinc.logger.debug(
                f"CLIInstance:analyse -> output fields: " f"{[f[0:2] for f in fields]}"
            )

            methods = {}
            if interface.get("has_output_item", True):
                methods["__str__"] = (
                    lambda myself: myself.__repr__()
                    if myself._output_item == ""
                    else myself._output_item
                )
            if self._checker:
                methods["check"] = lambda myself: myself._checker

            self.output_type = make_dataclass(
                "Solution",
                fields,
                bases=(_GeneratedSolution,),
                namespace=methods,
                frozen=True,
            )

    def _reset_analysis(self):
        self._method = None

    async def solutions(
        self,
        timeout: Optional[timedelta] = None,
        nr_solutions: Optional[int] = None,
        processes: Optional[int] = None,
        random_seed: Optional[int] = None,
        all_solutions=False,
        intermediate_solutions=False,
        free_search: bool = False,
        optimisation_level: Optional[int] = None,
        verbose: bool = False,
        debug_output: Optional[Path] = None,
        **kwargs,
    ):
        method = self.method  # Ensure self.analyse() has been executed
        # Set standard command line arguments
        cmd: List[Any] = [
            "--output-mode",
            "json",
            "--output-time",
            "--output-objective",
            "--output-output-item",
            "--statistics",  # Enable statistics
            # Enable intermediate solutions
            # (ensure that solvers always output their best solution)
            "--intermediate-solutions",
        ]

        # Process number of solutions to be generated
        if all_solutions:
            if nr_solutions is not None:
                raise ValueError(
                    "The number of solutions cannot be limited when looking "
                    "for all solutions"
                )
            if method != Method.SATISFY:
                raise NotImplementedError(
                    "Finding all optimal solutions is not yet implemented"
                )
            if "-a" not in self._solver.stdFlags:
                raise NotImplementedError("Solver does not support the -a flag")
            cmd.append("--all-solutions")
        elif nr_solutions is not None:
            if nr_solutions <= 0:
                raise ValueError(
                    "The number of solutions can only be set to a positive "
                    "integer number"
                )
            if self.method != Method.SATISFY:
                raise NotImplementedError(
                    "Finding multiple optimal solutions is not yet implemented"
                )
            if "-n" not in self._solver.stdFlags:
                raise NotImplementedError("Solver does not support the -n flag")
            cmd.extend(["--num-solutions", str(nr_solutions)])
        # Set number of processes to be used
        if processes is not None:
            cmd.extend(["--parallel", str(processes)])
        # Set random seed to be used
        if random_seed is not None:
            cmd.extend(["--random-seed", str(random_seed)])
        # Enable free search if specified
        if free_search:
            cmd.append("--free-search")
        # Set compiler optimisation level if specified
        if optimisation_level is not None:
            cmd.extend(["-O", str(optimisation_level)])

        # Set time limit for the MiniZinc solving
        if timeout is not None:
            cmd.extend(["--time-limit", str(int(timeout.total_seconds() * 1000))])

        if verbose:
            cmd.append("--verbose")

        for flag, value in kwargs.items():
            if not flag.startswith("-"):
                flag = "--" + flag
            if type(value) is bool:
                if value:
                    cmd.append(flag)
            else:
                cmd.extend([flag, value])

        # Add files as last arguments
        with self.files() as files, self._solver.configuration() as solver:
            assert self.output_type is not None
            cmd.extend(files)

            status = Status.UNKNOWN
            last_status = Status.UNKNOWN
            code = 0
            statistics: Dict[str, Any] = {}

            try:
                # Run the MiniZinc process
                proc = await self._driver.create_process(cmd, solver=solver)
                assert isinstance(proc.stderr, asyncio.StreamReader)
                assert isinstance(proc.stdout, asyncio.StreamReader)

                # Python 3.7+: replace with asyncio.create_task
                read_stderr = asyncio.ensure_future(_read_all(proc.stderr))

                if self._driver.parsed_version >= (2, 6, 0):
                    async for obj in decode_async_json_stream(
                        proc.stdout, cls=MZNJSONDecoder, enum_map=self._enum_map
                    ):
                        solution, new_status, statistics = self._parse_stream_obj(
                            obj, statistics
                        )
                        if new_status is not None:
                            status = new_status
                        elif solution is not None:
                            if status == Status.UNKNOWN:
                                status = Status.SATISFIED
                            yield Result(status, solution, statistics)
                            last_status = status
                            solution = None
                            statistics = {}
                else:
                    async for raw_sol in _seperate_solutions(proc.stdout):
                        status = Status.SATISFIED
                        solution, statistics = parse_solution(
                            raw_sol,
                            self.output_type,
                            self._enum_map,
                            self._field_renames,
                        )
                        yield Result(Status.SATISFIED, solution, statistics)

                code = await proc.wait()
            except asyncio.IncompleteReadError as err:
                # End of Stream has been reached
                # Read remaining text in buffer
                code = await proc.wait()
                remainder = err.partial

                # Parse and output the remaining statistics and status messages
                if self._driver.parsed_version >= (2, 6, 0):
                    for obj in decode_json_stream(
                        remainder, cls=MZNJSONDecoder, enum_map=self._enum_map
                    ):
                        solution, new_status, statistics = self._parse_stream_obj(
                            obj, statistics
                        )
                        if new_status is not None:
                            status = new_status
                        elif solution is not None:
                            if status == Status.UNKNOWN:
                                status = Status.SATISFIED
                            yield Result(status, solution, statistics)
                            solution = None
                            statistics = {}
                else:
                    for res in filter(None, remainder.split(SEPARATOR)):
                        new_status = Status.from_output(res, method)
                        if new_status is not None:
                            status = new_status
                        solution, statistics = parse_solution(
                            res,
                            self.output_type,
                            self._enum_map,
                            self._field_renames,
                        )
                        yield Result(status, solution, statistics)
            except (asyncio.CancelledError, MiniZincError, Exception):
                # Process was cancelled by the user, a MiniZincError occurred, or
                # an unexpected Python exception occurred
                # First, terminate the process
                proc.terminate()
                _ = await proc.wait()
                # Then, reraise the error that occurred
                raise
            if self._driver.parsed_version >= (2, 6, 0) and (
                status != last_status or statistics != {}
            ):
                yield Result(status, None, statistics)

            # Raise error if required
            stderr = await read_stderr
            if code != 0 or status == Status.ERROR:
                raise parse_error(stderr)

            if debug_output is not None:
                debug_output.write_bytes(stderr)

    @contextlib.contextmanager
    def flat(
        self,
        timeout: Optional[timedelta] = None,
        optimisation_level: Optional[int] = None,
        **kwargs,
    ):
        """Produce a FlatZinc file for the instance.

        Args:
            timeout (Optional[timedelta]): Set the time limit for the process
                of flattening the instance. TODO: An exception is raised if the
                timeout is reached.
            optimisation_level (Optional[int]): Set the MiniZinc compiler
                optimisation level.

                - 0: Disable optimisation
                - 1: Single pass optimisation (default)
                - 2: Flatten twice to improve flattening decisions
                - 3: Perform root-node-propagation
                - 4: Probe bounds of all variables at the root node
                - 5: Probe values of all variables at the root node

            **kwargs: Other flags to be passed to the compiler. ``--`` can be
                omitted in the name of the flag. If the type of the flag is
                Boolean, then its value signifies its occurrence.

        Yields:
            Tuple containing the files of the FlatZinc model, the output model
            and a dictionary the statistics of flattening

        """
        cmd: List[Any] = ["--compile", "--statistics"]

        fzn = tempfile.NamedTemporaryFile(prefix="fzn_", suffix=".fzn", delete=False)
        cmd.extend(["--fzn", fzn.name])
        fzn.close()
        ozn = tempfile.NamedTemporaryFile(prefix="ozn_", suffix=".fzn", delete=False)
        cmd.extend(["--ozn", ozn.name])
        ozn.close()

        if timeout is not None:
            cmd.extend(["--time-limit", str(int(timeout.total_seconds() * 1000))])

        # Set compiler optimisation level if specified
        if optimisation_level is not None:
            cmd.extend(["-O", str(optimisation_level)])

        for flag, value in kwargs.items():
            if not flag.startswith("-"):
                flag = "--" + flag
            if type(value) is bool:
                if value:
                    cmd.append(flag)
            else:
                cmd.extend([flag, value])

        # Add files as last arguments
        with self.files() as files:
            cmd.extend(files)
            # Run the MiniZinc process
            output = self._driver.run(cmd, solver=self._solver)

        statistics: Dict[str, Any] = {}
        matches = re.findall(rb"%%%mzn-stat:? (\w*)=([^\r\n]*)", output.stdout)
        for m in matches:
            set_stat(statistics, m[0].decode(), m[1].decode())

        try:
            yield fzn, ozn, statistics
        finally:
            os.remove(fzn.name)
            os.remove(ozn.name)

    def add_file(self, file: ParPath, parse_data: bool = True) -> None:
        self._reset_analysis()
        return super().add_file(file, parse_data)

    def add_string(self, code: str) -> None:
        self._reset_analysis()
        return super().add_string(code)

    def _parse_stream_obj(self, obj, statistics):
        solution = None
        status = None
        if obj["type"] == "solution":
            tmp = obj["output"]["json"]
            if "_objective" in tmp:
                tmp["objective"] = tmp.pop("_objective")
            if "_output" in tmp:
                tmp["_output_item"] = tmp.pop("_output")
            for before, after in self._field_renames:
                tmp[after] = tmp.pop(before)

            if "_checker" in statistics:
                tmp["_checker"] = statistics.pop("_checker")

            solution = self.output_type(**tmp)
            statistics["time"] = obj["time"]
        elif obj["type"] == "time":
            statistics["time"] = obj["time"]
        elif obj["type"] == "statistics":
            statistics.update(obj["statistics"])
        elif obj["type"] == "status":
            status = Status.from_str(obj["status"])
        elif obj["type"] == "checker":
            if "raw" in obj["output"]:
                statistics["_checker"] = obj["output"]["raw"]
            else:
                # TODO: can we ensure this is made JSON?
                statistics["_checker"] = obj["output"]["dzn"]

        return solution, status, statistics


async def _seperate_solutions(stream: asyncio.StreamReader):
    solution: bytes = b""
    while not stream.at_eof():
        try:
            solution += await stream.readuntil(SEPARATOR)
            yield solution
            solution = b""
        except asyncio.LimitOverrunError as err:
            solution += await stream.readexactly(err.consumed)


async def _read_all(stream: asyncio.StreamReader):
    output: bytes = b""
    while not stream.at_eof():
        try:
            output += await stream.read()
            return output
        except asyncio.LimitOverrunError as err:
            output += await stream.readexactly(err.consumed)
    return output



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from minizinc import Solver


def test_gecode():
    gecode = Solver.lookup("gecode")
    assert gecode is not None
    assert gecode.id.endswith("gecode")
    assert gecode.executable.endswith("fzn-gecode")


def test_chuffed():
    chuffed = Solver.lookup("chuffed")
    assert chuffed is not None
    assert chuffed.id.endswith("chuffed")
    assert chuffed.executable.endswith("fzn-chuffed")


def test_coinbc():
    coinbc = Solver.lookup("coin-bc")
    assert coinbc is not None
    assert coinbc.id.endswith("coin-bc")
    assert coinbc.executable is None



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import re
import subprocess
import sys
import warnings
from asyncio import create_subprocess_exec
from asyncio.subprocess import PIPE, Process
from dataclasses import fields
from json import loads
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union

import minizinc

from ..driver import Driver
from ..error import ConfigurationError, parse_error
from ..json import decode_json_stream
from ..solver import Solver

#: MiniZinc version required by the python package
CLI_REQUIRED_VERSION = (2, 5, 0)


def to_python_type(mzn_type: dict) -> Type:
    """Converts MiniZinc JSON type to Type

    Converts a MiniZinc JSON type definition generated by the MiniZinc CLI to a
    Python Type object. This can be used on types that result from calling
    ``minizinc --model-interface-only``.

    Args:
        mzn_type (dict): MiniZinc type definition as resulting from JSON

    Returns:
        Type: Type definition in Python

    """
    basetype = mzn_type["type"]
    pytype: Type
    # TODO: MiniZinc does not report enumerated types correctly
    if basetype == "bool":
        pytype = bool
    elif basetype == "float":
        pytype = float
    elif basetype == "int":
        pytype = int
    elif basetype == "string":
        pytype = str
    elif basetype == "ann":
        pytype = str
    else:
        warnings.warn(
            f"Unable to determine minizinc type `{basetype}` assuming integer type",
            FutureWarning,
        )
        pytype = int

    if mzn_type.get("set", False):
        if pytype is int:
            pytype = Union[Set[int], range]  # type: ignore
        else:
            pytype = Set[pytype]  # type: ignore

    dim = mzn_type.get("dim", 0)
    while dim >= 1:
        # No typing support for n-dimensional typing
        pytype = List[pytype]  # type: ignore
        dim -= 1
    return pytype


class CLIDriver(Driver):
    """Driver that interfaces with MiniZinc through the command line interface.

    The command line driver will interact with MiniZinc and its solvers through
    the use of a ``minizinc`` executable. Driving MiniZinc using its executable
    is non-incremental and can often trigger full recompilation and might
    restart the solver from the beginning when changes are made to the instance.

    Attributes:
        executable (Path): The path to the executable used to access the MiniZinc Driver

    """

    _executable: Path
    _solver_cache: Optional[Dict[str, Solver]] = None
    _version: Optional[Tuple[int, ...]] = None

    def __init__(self, executable: Path):
        self._executable = executable
        assert self._executable.exists()

        super(CLIDriver, self).__init__()

        if self.parsed_version < CLI_REQUIRED_VERSION:
            raise ConfigurationError(
                f"The MiniZinc driver found at '{self._executable}' has "
                f"version {self.parsed_version}. The minimal required version is "
                f"{CLI_REQUIRED_VERSION}."
            )

    def make_default(self) -> None:
        from . import CLIInstance

        minizinc.default_driver = self
        minizinc.Instance = CLIInstance

    def run(
        self,
        args: List[Any],
        solver: Optional[Solver] = None,
    ):
        # TODO: Add documentation
        windows_spawn_options: Dict[str, Any] = {}
        if sys.platform == "win32":
            # On Windows, MiniZinc terminates its subprocesses by generating a
            # Ctrl+C event for its own console using GenerateConsoleCtrlEvent.
            # Therefore, we must spawn it in its own console to avoid receiving
            # that Ctrl+C ourselves.
            #
            # On POSIX systems, MiniZinc terminates its subprocesses by sending
            # SIGTERM to the solver's process group, so this workaround is not
            # necessary as we won't receive that signal.
            windows_spawn_options = {
                "startupinfo": subprocess.STARTUPINFO(
                    dwFlags=subprocess.STARTF_USESHOWWINDOW,
                    wShowWindow=subprocess.SW_HIDE,
                ),
                "creationflags": subprocess.CREATE_NEW_CONSOLE,
            }

        # TODO: Always add --json-stream once 2.6.0 is minimum requirement
        if self.parsed_version >= (2, 6, 0):
            args.append("--json-stream")

        if solver is None:
            cmd = [str(self._executable), "--allow-multiple-assignments"] + [
                str(arg) for arg in args
            ]
            minizinc.logger.debug(f"CLIDriver:run -> command: \"{' '.join(cmd)}\"")
            output = subprocess.run(
                cmd,
                stdin=None,
                stdout=PIPE,
                stderr=PIPE,
                **windows_spawn_options,
            )
        else:
            with solver.configuration() as conf:
                cmd = [
                    str(self._executable),
                    "--solver",
                    conf,
                    "--allow-multiple-assignments",
                ] + [str(arg) for arg in args]
                minizinc.logger.debug(f"CLIDriver:run -> command: \"{' '.join(cmd)}\"")
                output = subprocess.run(
                    cmd,
                    stdin=None,
                    stdout=PIPE,
                    stderr=PIPE,
                    **windows_spawn_options,
                )
        if output.returncode != 0:
            if self.parsed_version >= (2, 6, 0):
                # Error will (usually) be raised in json stream
                for _ in decode_json_stream(output.stdout):
                    pass
            raise parse_error(output.stderr)
        return output

    async def create_process(
        self, args: List[str], solver: Optional[str] = None
    ) -> Process:
        """Start an asynchronous driver process with given arguments

        Args:
            args (List[str]): direct arguments to the driver
            solver (Union[str, Path, None]): Solver configuration string
                guaranteed by the user to be valid until the process has ended.
        """

        windows_spawn_options: Dict[str, Any] = {}
        if sys.platform == "win32":
            # See corresponding comment in run()
            windows_spawn_options = {
                "startupinfo": subprocess.STARTUPINFO(
                    dwFlags=subprocess.STARTF_USESHOWWINDOW,
                    wShowWindow=subprocess.SW_HIDE,
                ),
                "creationflags": subprocess.CREATE_NEW_CONSOLE,
            }

        # TODO: Always add --json-stream once 2.6.0 is minimum requirement
        if self.parsed_version >= (2, 6, 0):
            args.append("--json-stream")

        if solver is None:
            minizinc.logger.debug(
                f"CLIDriver:create_process -> program: {str(self._executable)} "
                f'args: "--allow-multiple-assignments '
                f"{' '.join(str(arg) for arg in args)}\""
            )
            proc = await create_subprocess_exec(
                str(self._executable),
                "--allow-multiple-assignments",
                *[str(arg) for arg in args],
                stdin=None,
                stdout=PIPE,
                stderr=PIPE,
                **windows_spawn_options,
            )
        else:
            minizinc.logger.debug(
                f"CLIDriver:create_process -> program: {str(self._executable)} "
                f'args: "--solver {solver} --allow-multiple-assignments '
                f"{' '.join(str(arg) for arg in args)}\""
            )
            proc = await create_subprocess_exec(
                str(self._executable),
                "--solver",
                solver,
                "--allow-multiple-assignments",
                *[str(arg) for arg in args],
                stdin=None,
                stdout=PIPE,
                stderr=PIPE,
                **windows_spawn_options,
            )
        return proc

    @property
    def minizinc_version(self) -> str:
        return self.run(["--version"]).stdout.decode()

    @property
    def parsed_version(self) -> Tuple[int, ...]:
        if self._version is None:
            output = subprocess.run(
                [str(self._executable), "--version"],
                stdin=None,
                stdout=PIPE,
                stderr=PIPE,
            )
            match = re.search(rb"version (\d+)\.(\d+)\.(\d+)", output.stdout)
            assert match
            self._version = tuple([int(i) for i in match.groups()])
        return self._version

    def available_solvers(self, refresh=False):
        if not refresh and self._solver_cache is not None:
            return self._solver_cache

        # Find all available solvers
        output = self.run(["--solvers-json"])
        solvers = loads(output.stdout)

        # Construct Solver objects
        self._solver_cache = {}
        allowed_fields = set([f.name for f in fields(Solver)])
        for s in solvers:
            obj = Solver(
                **{key: value for (key, value) in s.items() if key in allowed_fields}
            )
            if obj.version == "<unknown version>":
                obj._identifier = obj.id
            else:
                obj._identifier = obj.id + "@" + obj.version

            names = s.get("tags", [])
            names.extend([s["id"], s["id"].split(".")[-1]])
            for name in names:
                self._solver_cache.setdefault(name, []).append(obj)

        return self._solver_cache



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import tempfile
from dataclasses import fields

from support import InstanceTestCase

from minizinc.instance import Method
from minizinc.result import Status


class TestSatisfy(InstanceTestCase):
    code = "var 1..5: x"

    def test_solve(self):
        assert self.instance.method == Method.SATISFY
        result = self.instance.solve()
        assert result.status == Status.SATISFIED
        assert result["x"] in range(1, 5 + 1)

    def test_all_solution(self):
        result = self.instance.solve(all_solutions=True)
        assert result.status == Status.ALL_SOLUTIONS
        assert len(result) == 5
        assert sorted([sol.x for sol in result.solution]) == [
            i for i in range(1, 5 + 1)
        ]

    def test_nr_solutions(self):
        result = self.instance.solve(nr_solutions=3)
        assert result.status == Status.SATISFIED
        assert len(result) == 3
        for sol in result.solution:
            assert sol.x in range(1, 5 + 1)


class TestChecker(InstanceTestCase):
    code = "var 1..5: x"

    def test_checker(self):
        try:
            checker = tempfile.NamedTemporaryFile(
                prefix="checker_test_", suffix=".mzc.mzn", delete=False
            )
            checker.write(b'output["SIMPLE CHECK"];\n')
            checker.close()
            self.instance.add_file(checker.name)
            assert self.instance.method == Method.SATISFY
            result = self.instance.solve()
            assert result.status == Status.SATISFIED
            assert result["x"] in range(1, 5 + 1)
            assert result.solution.check().strip() == "SIMPLE CHECK"
        finally:
            os.remove(checker.name)

    def test_no_output_checker(self):
        try:
            checker = tempfile.NamedTemporaryFile(
                prefix="checker_test_", suffix=".mzc.mzn", delete=False
            )
            checker.write(b"int: data :: add_to_output = 2;\n")
            checker.close()
            self.instance.add_file(checker.name)
            assert self.instance.method == Method.SATISFY
            result = self.instance.solve()
            assert result.status == Status.SATISFIED
            assert result["x"] in range(1, 5 + 1)
            assert result.solution.check().strip() == "data = 2;"
        finally:
            os.remove(checker.name)


class TestMaximise(InstanceTestCase):
    code = """
        array[1..5] of var 1..5: x;
        solve ::int_search(x, input_order, indomain_min) maximize sum(x);
    """

    def test_solve(self):
        assert self.instance.method == Method.MAXIMIZE
        result = self.instance.solve()
        assert result.status == Status.OPTIMAL_SOLUTION
        assert result.objective == 25

    def test_intermediate(self):
        result = self.instance.solve(intermediate_solutions=True)
        assert len(result) == 21
        assert result.objective == 25


class TestParameter(InstanceTestCase):
    code = """
        int: n; % The number of queens.
        array [1..n] of var 1..n: q;

        include "alldifferent.mzn";

        constraint alldifferent(q);
        constraint alldifferent(i in 1..n)(q[i] + i);
        constraint alldifferent(i in 1..n)(q[i] - i);
    """

    def test_2(self):
        self.instance["n"] = 2
        assert self.instance.method == Method.SATISFY
        result = self.instance.solve()
        assert result.status == Status.UNSATISFIABLE

    def test_4(self):
        self.instance["n"] = 4
        assert self.instance.method == Method.SATISFY
        result = self.instance.solve()
        assert result.status == Status.SATISFIED
        assert len(result["q"]) == 4


class CheckEmpty(InstanceTestCase):
    code = """int: x = 5;"""

    def test_empty(self):
        assert self.instance.method == Method.SATISFY
        result = self.instance.solve()
        assert len(fields(result.solution)) == 0
        assert result.status == Status.SATISFIED



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
from pathlib import Path

from setuptools import find_packages, setup

setup(
    name="minizinc",
    use_scm_version={
        "write_to": "src/minizinc/_version.py",
        "write_to_template": 'version = "{version}"\n',
    },
    python_requires=">=3.6",
    author="Jip J. Dekker",
    author_email="jip.dekker@monash.edu",
    description="Access MiniZinc directly from Python",
    long_description=Path("README.md").read_text(encoding="UTF-8"),
    long_description_content_type="text/markdown",
    url="https://www.minizinc.org/",
    project_urls={
        "Bug Tracker": "https://github.com/MiniZinc/minizinc-python/issues",
        "Documentation": "https://minizinc-python.readthedocs.io",
        "Source": "https://github.com/MiniZinc/minizinc-python",
    },
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    install_requires=[
        "dataclasses>=0.6.0; python_version < '3.7'",
    ],
    extras_require={
        "dzn": ["lark-parser>=0.7.5"],
    },
    entry_points="""
        [pygments.lexers]
        minizinclexer = minizinc.pygments:MiniZincLexer
    """,
)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest

from minizinc.model import UnknownExpression

lark = pytest.importorskip("lark")
from minizinc.dzn import parse_dzn  # noqa: E402


def test_dzn_empty():
    assert parse_dzn("") == {}


def test_dzn_int():
    assert parse_dzn("x = 0") == {"x": 0}
    assert parse_dzn("x = -10") == {"x": -10}
    assert parse_dzn("x = 2123") == {"x": 2123}
    assert parse_dzn("x = 0xFF") == {"x": 255}
    assert parse_dzn("x = -0x100") == {"x": -256}
    assert parse_dzn("x = 0x0") == {"x": 0}
    assert parse_dzn("x = 0o23") == {"x": 19}
    assert parse_dzn("x = -0o30") == {"x": -24}
    assert parse_dzn("x = 0o0") == {"x": 0}


def test_dzn_float():
    assert parse_dzn("x = 0.0") == {"x": 0.0}
    assert parse_dzn("x = 0.125") == {"x": 0.125}
    assert parse_dzn("x = -0.99") == {"x": -0.99}
    assert parse_dzn("x = 3.8e10") == {"x": 3.8e10}
    assert parse_dzn("x = -1.4e7") == {"x": -1.4e7}
    assert parse_dzn("x = 2.45E-3") == {"x": 2.45e-3}
    assert parse_dzn("x = -1.33e-2") == {"x": -1.33e-2}
    assert parse_dzn("x = 4e12") == {"x": 4e12}
    assert parse_dzn("x = -3E10") == {"x": -3e10}
    assert parse_dzn("x = 2e-110") == {"x": 2e-110}
    assert parse_dzn("x = -9e-124") == {"x": -9e-124}


def test_dzn_string():
    assert parse_dzn('x = ""') == {"x": ""}
    assert parse_dzn('x = "test string"') == {"x": "test string"}
    assert parse_dzn('x = "ކ"') == {"x": "ކ"}
    assert parse_dzn('x = "🐛"') == {"x": "🐛"}


def test_dzn_set():
    # Set literals
    assert parse_dzn("x = {}") == {"x": set()}
    assert parse_dzn("x = {1}") == {"x": {1}}
    assert parse_dzn("x = {1,2,3}") == {"x": {1, 2, 3}}
    assert parse_dzn("x = {1,1,2}") == {"x": {1, 1, 2}}
    assert parse_dzn("x = {1.2,2.1}") == {"x": {1.2, 2.1}}

    # Set Ranges
    # note: upper range limit is exclusive in Python
    assert parse_dzn("x = 1..1") == {"x": range(1, 2)}
    assert set(parse_dzn("x = 1..1")["x"]) == {1}
    assert parse_dzn("x = 1..3") == {"x": range(1, 4)}
    assert set(parse_dzn("x = 1..3")["x"]) == {1, 2, 3}


def test_dzn_array():
    assert parse_dzn("x = []") == {"x": []}
    assert parse_dzn("x = [1]") == {"x": [1]}
    assert parse_dzn("x = [1,2,3,4]") == {"x": [1, 2, 3, 4]}
    assert parse_dzn("x = [2.1]") == {"x": [2.1]}
    assert parse_dzn("x = [2.1,3.2,4.2]") == {"x": [2.1, 3.2, 4.2]}
    assert parse_dzn('x = ["str1", "str2"]') == {"x": ["str1", "str2"]}
    assert parse_dzn("x = [{1,2,3}, {1,2}]") == {"x": [{1, 2, 3}, {1, 2}]}


def test_dzn_array2d():
    assert parse_dzn("x = [||]") == {"x": []}
    assert parse_dzn("x = [|1|2|3|]") == {"x": [[1], [2], [3]]}
    assert parse_dzn("x = [|1,4|2,5|3,6|]") == {"x": [[1, 4], [2, 5], [3, 6]]}
    assert parse_dzn("x = [|1.1,4.4|2.2,5.5|3.3,6.6|]") == {
        "x": [[1.1, 4.4], [2.2, 5.5], [3.3, 6.6]]
    }


def test_dzn_unknown():
    assert parse_dzn("x = array2d(index1,1..2,[4, 3, 4, 5, 3, 6]);") == {
        "x": UnknownExpression("array2d(index1,1..2,[4, 3, 4, 5, 3, 6])")
    }


def test_dzn_semicolon():
    assert parse_dzn("x = 1;") == {"x": 1}
    assert parse_dzn("x = 1") == {"x": 1}
    assert parse_dzn("x = 1; y = 2") == {"x": 1, "y": 2}
    assert parse_dzn("x = 1; y = 2;") == {"x": 1, "y": 2}
    assert parse_dzn('x = -20; y = 2e3; z = "string"') == {
        "x": -20,
        "y": 2e3,
        "z": "string",
    }


def test_dzn_trailing_comma():
    assert parse_dzn("x = [1,2,3,]") == {"x": [1, 2, 3]}
    assert parse_dzn("x = {1,2,3,}") == {"x": {1, 2, 3}}
    assert parse_dzn("x = [{1,},{2,},]") == {"x": [{1}, {2}]}
    assert parse_dzn("x = [|1,|2,|3,|]") == {"x": [[1], [2], [3]]}
